from pathlib import Path

from poetry.console.commands.command import Command
from polylith import alias, info, project, repo, workspace
from polylith.libs import report
from polylith.poetry.commands.check import command_options
from polylith.poetry.internals import find_third_party_libs


class LibsCommand(Command):
    name = "poly libs"
    description = "Show third-party libraries used in the workspace."

    options = command_options

    def print_report(self, root: Path, ns: str, data: dict) -> bool:
        is_strict = self.option("strict")

        name = data["name"]
        path = data["path"]

        brick_imports = report.get_third_party_imports(root, ns, data)

        report.print_libs_summary(brick_imports, data)
        report.print_libs_in_bricks(brick_imports)

        try:
            third_party_libs = find_third_party_libs(self.poetry, path)

            library_aliases = alias.parse(self.option("alias"))
            extra = alias.pick(library_aliases, third_party_libs)

            libs = third_party_libs.union(extra)

            return report.print_missing_installed_libs(
                brick_imports,
                libs,
                name,
                is_strict,
            )
        except ValueError as e:
            self.line_error(f"{name}: <error>{e}</error>")
            return False

    def handle(self) -> int:
        root = repo.get_workspace_root(Path.cwd())
        ns = workspace.parser.get_namespace_from_config(root)

        projects_data = info.get_projects_data(root, ns)

        if self.option("directory"):
            project_name = project.get_project_name(self.poetry.pyproject.data)
            data = next((p for p in projects_data if p["name"] == project_name), None)

            if not data:
                raise ValueError(f"Didn't find project in {self.option('directory')}")

            res = self.print_report(root, ns, data)
            return 0 if res else 1

        results = {self.print_report(root, ns, data) for data in projects_data}

        return 0 if all(results) else 1



import difflib
from pathlib import Path
from typing import Set

from polylith import info, workspace
from polylith.libs import grouping
from polylith.reporting import theme
from rich import box
from rich.console import Console
from rich.padding import Padding
from rich.table import Table


def get_third_party_imports(root: Path, ns: str, project_data: dict) -> dict:
    bases = {b for b in project_data.get("bases", [])}
    components = {c for c in project_data.get("components", [])}

    bases_paths = workspace.paths.collect_bases_paths(root, ns, bases)
    components_paths = workspace.paths.collect_components_paths(root, ns, components)

    bases_imports = grouping.get_third_party_imports(root, bases_paths)
    components_imports = grouping.get_third_party_imports(root, components_paths)

    return {"bases": bases_imports, "components": components_imports}


def flatten_imports(brick_imports: dict, brick: str) -> Set[str]:
    return set().union(*brick_imports.get(brick, {}).values())


def flatten_brick_imports(brick_imports: dict) -> Set[str]:
    bases_imports = flatten_imports(brick_imports, "bases")
    components_imports = flatten_imports(brick_imports, "components")

    return set().union(bases_imports, components_imports)


def filter_close_matches(
    unknown_imports: Set[str], dependencies: Set[str], cutoff: float
) -> Set[str]:
    unknowns = {str.lower(u) for u in unknown_imports}
    deps = {str.lower(d) for d in dependencies}

    return {
        u for u in unknowns if not difflib.get_close_matches(u, deps, cutoff=cutoff)
    }


def get_unknowns(brick_imports: dict, deps: Set[str]) -> Set[str]:
    imports = flatten_brick_imports(brick_imports)

    return imports.difference(deps)


def calculate_diff(
    brick_imports: dict, deps: Set[str], is_strict: bool = False
) -> Set[str]:
    unknown_imports = get_unknowns(brick_imports, deps)

    cutoff = 0.6 if not is_strict else 1

    return filter_close_matches(unknown_imports, deps, cutoff)


def print_libs_summary(brick_imports: dict, project_data: dict) -> None:
    console = Console(theme=theme.poly_theme)

    name = project_data["name"]
    is_project = info.is_project(project_data)

    printable_name = f"[proj]{name}[/]" if is_project else "[data]development[/]"
    console.print(
        Padding(f"[data]Libraries summary for [/]{printable_name}", (1, 0, 1, 0))
    )

    bases_imports = flatten_imports(brick_imports, "bases")
    components_imports = flatten_imports(brick_imports, "components")

    console.print(f"[base]Libraries used in bases[/]: [data]{len(bases_imports)}[/]")
    console.print(
        f"[comp]libraries used in components[/]: [data]{len(components_imports)}[/]"
    )


def print_libs_in_bricks(brick_imports: dict) -> None:
    bases_imports = flatten_imports(brick_imports, "bases")
    components_imports = flatten_imports(brick_imports, "components")

    if not bases_imports and not components_imports:
        return

    console = Console(theme=theme.poly_theme)
    table = Table(box=box.SIMPLE_HEAD)

    bases = brick_imports.get("bases", {})
    components = brick_imports.get("components", {})

    table.add_column("[data]brick[/]")
    table.add_column("[data]libraries[/]")

    for brick, imports in bases.items():
        table.add_row(f"[base]{brick}[/]", ", ".join(sorted(imports)))

    for brick, imports in components.items():
        table.add_row(f"[comp]{brick}[/]", ", ".join(sorted(imports)))

    console.print(table, overflow="ellipsis")


def print_missing_installed_libs(
    brick_imports: dict, third_party_libs: Set[str], project_name: str, is_strict: bool = False
) -> bool:
    diff = calculate_diff(brick_imports, third_party_libs, is_strict)

    if not diff:
        return True

    console = Console(theme=theme.poly_theme)

    missing = ", ".join(sorted(diff))

    console.print(
        f"[data]Could not locate all libraries in [/][proj]{project_name}[/]. [data]Caused by missing dependencies?[/]"
    )

    console.print(f":thinking_face: {missing}")
    return False



from pathlib import Path
from typing import List, Union

from polylith.bricks import component
from polylith.bricks.brick import create_brick
from polylith.repo import bases_dir
from polylith.test import create_test


def create_base(path: Path, namespace: str, package: str, description: Union[str, None]) -> None:
    create_brick(path, bases_dir, namespace, package, description)
    create_test(path, bases_dir, namespace, package)


def get_bases_data(path: Path, ns: str) -> List[dict]:
    return component.get_components_data(path, ns, bases_dir)



from pathlib import Path
from typing import List

from poetry.console.commands.command import Command
from polylith import info, project, repo, sync, workspace


class SyncCommand(Command):
    name = "poly sync"
    description = "Update <comment>pyproject.toml</comment> with missing bricks."

    def filter_projects_data(self, all_projects_data: List[dict]) -> List[dict]:
        if self.option("directory"):
            project_name = project.get_project_name(self.poetry.pyproject.data)

            data = next(
                (p for p in all_projects_data if p["name"] == project_name), None
            )

            if not data:
                raise ValueError(f"Didn't find project in {self.option('directory')}")

            return [data]

        return all_projects_data

    def handle(self) -> int:
        is_verbose = self.option("verbose")
        is_quiet = self.option("quiet")

        root = repo.get_workspace_root(Path.cwd())
        ns = workspace.parser.get_namespace_from_config(root)

        bases = info.get_bases(root, ns)
        components = info.get_components(root, ns)
        workspace_data = {"bases": bases, "components": components}

        all_projects_data = info.get_bricks_in_projects(root, components, bases, ns)
        projects_data = self.filter_projects_data(all_projects_data)

        diffs = [
            sync.calculate_diff(root, ns, data, workspace_data)
            for data in projects_data
        ]

        for diff in diffs:
            sync.update_project(root, ns, diff)

            if is_quiet:
                continue

            sync.report.print_summary(diff)

            if is_verbose:
                sync.report.print_brick_imports(diff)

        return 0



from polylith.libs import report
from polylith.libs.grouping import extract_third_party_imports, get_third_party_imports

__all__ = ["report", "extract_third_party_imports", "get_third_party_imports"]



from functools import reduce
from pathlib import Path
from typing import Dict, Iterable, List, Set, Union

from poetry.factory import Factory
from poetry.poetry import Poetry
from poetry.utils.env import EnvManager


def top_level_packages(dist) -> List[str]:
    top_level = dist.read_text("top_level.txt")

    return str.split(top_level or "")


def mapped_packages(dist) -> dict:
    packages = top_level_packages(dist)
    name = dist.metadata["name"]

    return {name: packages} if packages else {}


def map_packages(acc, dist) -> dict:
    return {**acc, **mapped_packages(dist)}


def distributions_packages(dists) -> Dict[str, List[str]]:
    """Return a mapping of top-level packages to their distributions."""
    return reduce(map_packages, dists, {})


def get_project_poetry(poetry: Poetry, path: Union[Path, None]) -> Poetry:
    return Factory().create_poetry(path) if path else poetry


def distributions(poetry: Poetry, path: Union[Path, None]) -> Iterable:
    project_poetry = get_project_poetry(poetry, path)

    env = EnvManager(project_poetry).get()

    return env.site_packages.distributions()


def find_third_party_libs(poetry: Poetry, path: Union[Path, None]) -> Set:
    project_poetry = get_project_poetry(poetry, path)

    if not project_poetry.locker.is_locked():
        raise ValueError("poetry.lock not found. Run `poetry lock` to create it.")

    packages = project_poetry.locker.locked_repository().packages

    return {p.name for p in packages}



from pathlib import Path
from typing import List, Set

from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith import diff, info, repo, workspace


class DiffCommand(Command):
    name = "poly diff"
    description = "Shows changed bricks compared to the latest git tag."

    options = [
        option(
            long_name="short",
            short_name="s",
            description="Print short view",
            flag=True,
        ),
        option(
            long_name="bricks",
            description="Print changed bricks",
            flag=True,
        ),
        option(
            long_name="since",
            description="Changed since a specific tag",
            flag=False,
        ),
    ]

    def has_partial_options(self) -> bool:
        return any(self.option(k) for k in {"bricks"})

    def print_partial_views(
        self,
        affected_projects: Set[str],
        bases: List[str],
        components: List[str],
    ) -> None:
        short = self.option("short")

        if short and not self.has_partial_options():
            diff.report.print_projects_affected_by_changes(affected_projects, short)

            return

        if self.option("bricks"):
            diff.report.print_detected_changes_in_bricks(bases, components, short)

    def print_views(self, root: Path, tag: str) -> None:
        ns = workspace.parser.get_namespace_from_config(root)
        files = diff.collect.get_files(tag)
        bases = diff.collect.get_changed_bases(files, ns)
        components = diff.collect.get_changed_components(files, ns)
        projects = diff.collect.get_changed_projects(files)
        all_projects_data = info.get_bricks_in_projects(root, components, bases, ns)
        projects_data = [p for p in all_projects_data if info.is_project(p)]

        affected_projects = diff.collect.get_projects_affected_by_changes(
            projects_data, projects, bases, components
        )

        short = self.option("short")

        if not short and not self.has_partial_options():
            diff.report.print_diff_summary(tag, bases, components)
            diff.report.print_detected_changes_in_projects(projects, short)
            diff.report.print_diff_details(projects_data, bases, components)

            return

        self.print_partial_views(affected_projects, bases, components)

    def handle(self) -> int:
        root = repo.get_workspace_root(Path.cwd())

        tag_name = self.option("since")
        tag = diff.collect.get_latest_tag(root, tag_name)

        if not tag:
            self.line("No tags found in repository.")
        else:
            self.print_views(root, tag)

        return 0



from pathlib import Path
from typing import Union

from polylith.dirs import create_dir
from polylith.files import create_file
from polylith.interface import create_interface
from polylith.readme import create_brick_readme
from polylith.workspace import parser


def create_brick(
    root: Path,
    brick: str,
    namespace: str,
    package: str,
    description: Union[str, None],
    modulename: str = "core",
) -> None:
    path_kwargs = {"brick": brick, "namespace": namespace, "package": package}

    brick_structure = parser.get_brick_structure_from_config(root)
    resources_structure = parser.get_resources_structure_from_config(root)

    brick_path = brick_structure.format(**path_kwargs)
    resources_path = resources_structure.format(**path_kwargs)

    d = create_dir(root, brick_path)
    create_file(d, f"{modulename}.py")
    create_interface(d, namespace, package, modulename, description)

    if parser.is_readme_generation_enabled(root):
        create_brick_readme(root / resources_path, package, brick, description)



from pathlib import Path

from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith import alias, check, info, project, repo, workspace
from polylith.poetry import internals

command_options = [
    option(
        long_name="strict",
        description="More strict checks when matching name of third-party libraries and imports",
        flag=True,
    ),
    option(
        long_name="alias",
        description="alias for a third-party library, useful when an import differ from the library name",
        flag=False,
        multiple=True,
    ),
]


class CheckCommand(Command):
    name = "poly check"
    description = "Validates the <comment>Polylith</> workspace."

    options = command_options

    def print_report(self, root: Path, ns: str, project_data: dict) -> bool:
        is_verbose = self.option("verbose")
        is_quiet = self.option("quiet")
        is_strict = self.option("strict")

        path = project_data["path"]
        name = project_data["name"]

        try:
            collected_imports = check.report.collect_all_imports(root, ns, project_data)
            third_party_libs = internals.find_third_party_libs(self.poetry, path)

            dists = internals.distributions(self.poetry, path)
            known_aliases = internals.distributions_packages(dists)
            known_aliases.update(alias.parse(self.option("alias")))

            extra = alias.pick(known_aliases, third_party_libs)

            libs = third_party_libs.union(extra)

            details = check.report.create_report(
                project_data,
                collected_imports,
                libs,
                is_strict,
            )

            res = all([not details["brick_diff"], not details["libs_diff"]])

            if is_quiet:
                return res

            check.report.print_missing_deps(details["brick_diff"], name)
            check.report.print_missing_deps(details["libs_diff"], name)

            if is_verbose:
                check.report.print_brick_imports(details["brick_imports"])
                check.report.print_brick_imports(details["third_party_imports"])

            return res
        except ValueError as e:
            self.line_error(f"{name}: <error>{e}</error>")
            return False

    def handle(self) -> int:
        root = repo.get_workspace_root(Path.cwd())
        ns = workspace.parser.get_namespace_from_config(root)

        all_projects_data = info.get_projects_data(root, ns)
        projects_data = [p for p in all_projects_data if info.is_project(p)]

        if self.option("directory"):
            project_name = project.get_project_name(self.poetry.pyproject.data)
            data = next((p for p in projects_data if p["name"] == project_name), None)

            if not data:
                raise ValueError(f"Didn't find project in {self.option('directory')}")

            res = self.print_report(root, ns, data)
            return 0 if res else 1

        results = {self.print_report(root, ns, data) for data in projects_data}

        return 0 if all(results) else 1



from pathlib import Path
from typing import List

import tomlkit
from polylith.repo import default_toml


def get_project_package_includes(data) -> List[dict]:
    return data["tool"]["poetry"].get("packages", [])


def get_project_name(data) -> str:
    return data["tool"]["poetry"]["name"]


def get_toml(path: Path) -> tomlkit.TOMLDocument:
    with path.open(encoding="utf-8", errors="ignore") as f:
        return tomlkit.loads(f.read())


def get_project_files(root: Path) -> dict:
    projects = sorted(root.glob(f"projects/**/{default_toml}"))
    development = Path(root / default_toml)

    proj = {"projects": projects}
    dev = {"development": [development]}

    return {**proj, **dev}


def toml_data(path: Path, project_type: str) -> dict:
    return {"toml": get_toml(path), "path": path.parent, "type": project_type}


def get_toml_files(root: Path) -> List[dict]:
    project_files = get_project_files(root)

    proj = [toml_data(p, "project") for p in project_files["projects"]]
    dev = [toml_data(d, "development") for d in project_files["development"]]

    return proj + dev


def get_packages_for_projects(root: Path) -> List[dict]:
    tomls = get_toml_files(root)

    return [
        {
            "name": get_project_name(d["toml"]),
            "packages": get_project_package_includes(d["toml"]),
            "path": d["path"],
            "type": d["type"],
        }
        for d in tomls
    ]



from pathlib import Path

from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith import info, repo, workspace


class InfoCommand(Command):
    name = "poly info"
    description = "Info about the <comment>Polylith</> workspace."

    options = [
        option(
            long_name="short",
            short_name="s",
            description="Display Workspace Info adjusted for many projects",
            flag=True,
        ),
    ]

    def handle(self) -> int:
        short = self.option("short")

        root = repo.get_workspace_root(Path.cwd())
        ns = workspace.parser.get_namespace_from_config(root)
        bases = info.get_bases(root, ns)
        components = info.get_components(root, ns)
        projects_data = info.get_bricks_in_projects(root, components, bases, ns)

        info.print_workspace_summary(projects_data, bases, components)

        if components or bases:
            if short:
                info.print_compressed_view_for_bricks_in_projects(
                    projects_data, bases, components
                )
            else:
                info.print_bricks_in_projects(projects_data, bases, components)

        return 0



from pathlib import Path
from typing import Union

import tomlkit
from polylith import repo
from polylith.dirs import create_dir
from polylith.repo import projects_dir

template = """\
[tool.poetry]
name = "{name}"
version = "0.1.0"
description = "{description}"
authors = {authors}
license = ""

packages = []

[tool.poetry.dependencies]
python = "{python_version}"

[tool.poetry.group.dev.dependencies]

[build-system]
requires = ["poetry-core>=1.0.0"]
build-backend = "poetry.core.masonry.api"
"""


def get_workspace_toml(path: Path) -> dict:
    with open(str(path / repo.default_toml), "r", errors="ignore") as f:
        data: dict = tomlkit.loads(f.read())

    return data


def create_project_toml(
    name: str, template: str, workspace_toml: dict, description: str
) -> tomlkit.TOMLDocument:
    authors = workspace_toml["tool"]["poetry"]["authors"]
    python_version = workspace_toml["tool"]["poetry"]["dependencies"]["python"]

    content = template.format(
        name=name,
        description=description,
        authors=authors,
        python_version=python_version,
    )

    return tomlkit.loads(content)


def create_project(
    path: Path, _namespace: str, name: str, description: Union[str, None]
) -> None:
    d = create_dir(path, f"{projects_dir}/{name}")

    workspace_toml = get_workspace_toml(path)
    project_toml = create_project_toml(
        name, template, workspace_toml, description or ""
    )

    fullpath = d / repo.default_toml

    with fullpath.open("w", encoding="utf-8") as f:
        f.write(tomlkit.dumps(project_toml))



from pathlib import Path

from polylith import repo, workspace


def create(command, fn):
    root = repo.get_workspace_root(Path.cwd())
    name = command.option("name")
    description = command.option("description")
    namespace = workspace.parser.get_namespace_from_config(root)

    if not name:
        raise ValueError("Please add a name by using --name or -n")

    if not namespace:
        raise ValueError(
            "Didn't find a namespace. Expected to find it in workspace.toml."
        )

    fn(root, namespace, name, description)



from pathlib import Path
from typing import List

import tomlkit
from polylith import project, repo, workspace
from tomlkit.toml_document import TOMLDocument


def to_package(namespace: str, brick: str, brick_path: str, theme: str) -> dict:
    folder = f"{brick_path}" if theme == "loose" else f"{brick_path}/{brick}/src"

    return {"include": f"{namespace}/{brick}", "from": folder}


def generate_updated_project(data: TOMLDocument, packages: List[dict]) -> str:
    original = tomlkit.dumps(data)
    copy: dict = tomlkit.parse(original)

    if copy["tool"]["poetry"].get("packages") is None:
        copy["tool"]["poetry"].add("packages", [])

    for package in packages:
        copy["tool"]["poetry"]["packages"].append(package)

    copy["tool"]["poetry"]["packages"].multiline(True)

    return tomlkit.dumps(copy)


def to_packages(root: Path, namespace: str, diff: dict) -> List[dict]:
    theme = workspace.parser.get_theme_from_config(root)

    is_project = diff["is_project"]

    bases_path = "../../bases" if is_project else "bases"
    components_path = "../../components" if is_project else "components"

    a = [to_package(namespace, b, bases_path, theme) for b in diff["bases"]]
    b = [to_package(namespace, c, components_path, theme) for c in diff["components"]]

    return a + b


def rewrite_project_file(path: Path, packages: List[dict]):
    fullpath = path / repo.default_toml
    project_toml = project.get_toml(fullpath)

    generated = generate_updated_project(project_toml, packages)

    with fullpath.open("w", encoding="utf-8") as f:
        f.write(generated)


def update_project(path: Path, namespace: str, diff: dict):
    packages = to_packages(path, namespace, diff)

    if packages:
        rewrite_project_file(diff["path"], packages)



from pathlib import Path

from polylith.dirs import create_dir
from polylith.files import create_file
from polylith.workspace import parser

template = """\
from {namespace}.{package} import {modulename}


def test_sample():
    assert {modulename} is not None
"""


def create_test(
    root: Path, brick: str, namespace: str, package: str, modulename: str = "core"
) -> None:
    if not parser.is_test_generation_enabled(root):
        return

    dirs_structure = parser.get_tests_structure_from_config(root)
    dirs = dirs_structure.format(brick=brick, namespace=namespace, package=package)
    d = create_dir(root, dirs)

    create_file(d, "__init__.py")
    test_file = create_file(d, f"test_{modulename}.py")

    content = template.format(
        namespace=namespace, package=package, modulename=modulename
    )

    test_file.write_text(content)
